from sklearn.datasets import load_iris
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import FloatRange, Categorical

# loading the iris dataset
X, y = load_iris(return_X_y=True)

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('multi_class_svm_pipe',
                    optimizer='random_grid_search',
                    optimizer_params={'n_configurations': 10},
                    metrics=['accuracy'],
                    best_config_metric='accuracy',
                    outer_cv=KFold(n_splits=3, shuffle=True),
                    inner_cv=KFold(n_splits=3, shuffle=True),
                    verbosity=1,
                    project_folder='./tmp/')


my_pipe.add(PipelineElement('StandardScaler'))

my_pipe += PipelineElement('SVC',
                           hyperparameters={'kernel': Categorical(['rbf', 'linear']),
                                            'C': FloatRange(0.5, 2)},
                           gamma='scale')

my_pipe.fit(X, y)



from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement, Switch
from photonai.optimization import FloatRange, IntegerRange

my_pipe = Hyperpipe('hp_switch_optimizer',
                    inner_cv=KFold(n_splits=5),
                    outer_cv=KFold(n_splits=3),
                    optimizer='switch',
                    optimizer_params={'name': 'sk_opt', 'n_configurations': 50},
                    metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                    best_config_metric='accuracy',
                    project_folder='./tmp',
                    verbosity=1)

my_pipe.add(PipelineElement('StandardScaler'))

my_pipe += PipelineElement('PCA',
                           hyperparameters={'n_components': IntegerRange(10, 30)},
                           test_disabled=True)

# set up two learning algorithms in an ensemble
estimator_selection = Switch('estimators')

estimator_selection += PipelineElement('RandomForestClassifier',
                                       criterion='gini',
                                       hyperparameters={'min_samples_split': IntegerRange(2, 4),
                                                        'max_features': ['auto', 'sqrt', 'log2'],
                                                        'bootstrap': [True, False]})
estimator_selection += PipelineElement('SVC',
                                       hyperparameters={'C': FloatRange(0.5, 25),
                                                        'kernel': ['linear', 'rbf']})

my_pipe += estimator_selection

X, y = load_breast_cancer(return_X_y=True)
my_pipe.fit(X, y)

my_pipe.results_handler.get_mean_of_best_validation_configs_per_estimator()



from sklearn.datasets import load_digits
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import Categorical

# WE USE THE BREAST CANCER SET FROM SKLEARN
X, y = load_digits(n_class=5, return_X_y=True)

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('basic_keras_multiclass_pipe',
                    optimizer='grid_search',
                    optimizer_params={},
                    metrics=['accuracy'],
                    best_config_metric='accuracy',
                    outer_cv=KFold(n_splits=2),
                    inner_cv=KFold(n_splits=2),
                    verbosity=1,
                    project_folder='./tmp/')


# ADD ELEMENTS TO YOUR PIPELINE
my_pipe.add(PipelineElement('StandardScaler'))

# attention: shape of hidden_layer_sizes == shape of activations. If you want to choose a function in every layer,
# grid_search eliminates combinations with len(hidden_layer_size) != len(activations).
# Check out: hidden_layer_sizes=[25, 10], activations=['tanh', 'relu']
my_pipe += PipelineElement('KerasDnnClassifier',
                           hyperparameters={'hidden_layer_sizes': Categorical([[20, 10, 5], [10, 8, 4]]),
                                            'dropout_rate': Categorical([0.5, [0.5, 0.5, 0.5]])},
                           activations='relu',
                           nn_batch_size=32,
                           epochs=50,
                           multi_class=True,
                           verbosity=0)

# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y)



import unittest
import numpy as np
import warnings
import glob
from shutil import rmtree

from sklearn.metrics import accuracy_score

from photonai.helper.helper import PhotonDataHelper
from photonai.base import PipelineElement, Switch, Hyperpipe
from photonai.base.photon_pipeline import PhotonPipeline
from sklearn.datasets import fetch_olivetti_faces
from sklearn.model_selection import ShuffleSplit
from photonai.optimization import FloatRange, Categorical, IntegerRange
from photonai.optimization.hyperparameters import NumberRange
from photonai.optimization.smac.smac import SMACOptimizer
import photonai.optimization.smac.smac as photonai_smac

if photonai_smac.__found__:
    # Import ConfigSpace and different types of parameters
    from smac.configspace import ConfigurationSpace
    from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
        UniformFloatHyperparameter, UniformIntegerHyperparameter
    from ConfigSpace.conditions import InCondition
    # Import SMAC-utilities
    from smac.scenario.scenario import Scenario
    from smac.facade.smac_bo_facade import SMAC4BO
    from smac.facade.smac_hpo_facade import SMAC4HPO
    from smac.facade.smac_ac_facade import SMAC4AC
    from smac.facade.smac_bohb_facade import BOHB4HPO


@unittest.skipIf(not photonai_smac.__found__, 'smac not available')
class Smac3IntegrationTest(unittest.TestCase):

    def setUp(self):
        self.s_split = ShuffleSplit(n_splits=3, test_size=0.2, random_state=42)

        self.time_limit = 20

        self.pipe = Hyperpipe('basic_svm_pipe',
                              optimizer='smac',
                              optimizer_params={'facade': SMAC4BO,
                                                'wallclock_limit': self.time_limit,
                                                'rng': 42},
                              metrics=['accuracy'],
                              random_seed=True,
                              best_config_metric='accuracy',
                              inner_cv=self.s_split,
                              verbosity=0,
                              project_folder='./tmp/')

    def simple_classification(self):
        dataset = fetch_olivetti_faces(download_if_missing=True)
        self.X = dataset["data"]
        self.y = dataset["target"]
        return self.X, self.y

    # integration test for simple pipeline without Switch
    def test_photon_implementation_simple(self):
        # PHOTON implementation
        self.pipe.add(PipelineElement('StandardScaler'))
        self.pipe += PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 30)})
        self.pipe += PipelineElement('SVC', hyperparameters={'kernel': Categorical(["rbf", 'poly']),
                                                             'C': FloatRange(0.5, 200)}, gamma='auto')
        self.X, self.y = self.simple_classification()
        self.pipe.fit(self.X, self.y)

        # direct AUTO ML implementation
        # Build Configuration Space which defines all parameters and their ranges
        cs = ConfigurationSpace()
        n_components = UniformIntegerHyperparameter("PCA__n_components", 5, 30)
        cs.add_hyperparameter(n_components)
        kernel = CategoricalHyperparameter("SVC__kernel", ["rbf", 'poly'])
        cs.add_hyperparameter(kernel)
        c = UniformFloatHyperparameter("SVC__C", 0.5, 200)
        cs.add_hyperparameter(c)

        # Scenario object
        scenario = Scenario({"run_obj": "quality",
                             "cs": cs,
                             "deterministic": "true",
                             "wallclock_limit": self.time_limit,
                             "limit_resources": False,
                             'abort_on_first_run_crash': False
                             })

        # Optimize, using a SMAC directly
        smac = SMAC4BO(scenario=scenario, rng=42, tae_runner=self.objective_function_simple)
        _ = smac.optimize()

        runhistory_photon = [x.metrics_test[0].value for x in self.pipe.results.outer_folds[0].tested_config_list]
        runhistory_original = [1-x for x in list(smac.solver.runhistory._cost_per_config.values())]

        min_len = min(len(runhistory_original), len(runhistory_photon))
        np.testing.assert_almost_equal(runhistory_photon[:min_len], runhistory_original[:min_len], 1)

    def objective_function_simple(self, cfg):
        cfg = {k: cfg[k] for k in cfg if cfg[k]}
        values = []

        train_indices = list(self.pipe.cross_validation.outer_folds.values())[0].train_indices
        self._validation_X, self._validation_y, _ = PhotonDataHelper.split_data(self.X, self.y,
                                                                                kwargs=None, indices=train_indices)

        for inner_fold in list(list(self.pipe.cross_validation.inner_folds.values())[0].values()):
            sc = PipelineElement("StandardScaler", {})
            pca = PipelineElement("PCA", {}, random_state=42)
            svc = PipelineElement("SVC", {}, random_state=42, gamma='auto')
            my_pipe = PhotonPipeline([('StandardScaler', sc), ('PCA', pca), ('SVC', svc)])
            my_pipe.set_params(**cfg)
            my_pipe.fit(self._validation_X[inner_fold.train_indices, :],
                        self._validation_y[inner_fold.train_indices])
            values.append(accuracy_score(self._validation_y[inner_fold.test_indices],
                                         my_pipe.predict(self._validation_X[inner_fold.test_indices, :])
                                         )
                          )
        return 1 - np.mean(values)

    def test_further_parameters(self):
        n_configurations = 4
        pipe = Hyperpipe('basic_svm_pipe',
                         optimizer='smac',
                         optimizer_params={'facade': SMAC4BO,
                                           'wallclock_limit': 1000,
                                           'ta_run_limit': n_configurations,
                                           'rng': 42},
                         metrics=['accuracy'],
                         random_seed=True,
                         best_config_metric='accuracy',
                         inner_cv=self.s_split,
                         verbosity=0,
                         project_folder='./tmp/')

        pipe.add(PipelineElement('StandardScaler'))
        pipe += PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 30)})
        pipe += PipelineElement('SVC', hyperparameters={'kernel': Categorical(["rbf", 'poly']),
                                                             'C': FloatRange(0.5, 200)}, gamma='auto')
        X, y = self.simple_classification()
        pipe.fit(X, y)
        self.assertEqual(len(pipe.results.outer_folds[0].tested_config_list), n_configurations)

    # integration test for pipeline with Switch
    def test_photon_implementation_switch(self):
        # PHOTON implementation
        self.pipe.add(PipelineElement('StandardScaler'))
        self.pipe += PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 30)})
        estimator_siwtch = Switch("Estimator")
        estimator_siwtch += PipelineElement('SVC', hyperparameters={'kernel': Categorical(["rbf", 'poly']),
                                                                    'C': FloatRange(0.5, 200)}, gamma='auto')
        estimator_siwtch += PipelineElement('RandomForestClassifier',
                                            hyperparameters={'criterion': Categorical(['gini', 'entropy']),
                                                             'min_samples_split': IntegerRange(2, 4)
                                                             })
        self.pipe += estimator_siwtch
        self.X, self.y = self.simple_classification()
        self.pipe.fit(self.X, self.y)

        # direct AUTO ML implementation

        # Build Configuration Space which defines all parameters and their ranges
        cs = ConfigurationSpace()
        n_components = UniformIntegerHyperparameter("PCA__n_components", 5, 30)
        cs.add_hyperparameter(n_components)

        switch = CategoricalHyperparameter("Estimator_switch", ['svc', 'rf'])
        cs.add_hyperparameter(switch)

        kernel = CategoricalHyperparameter("SVC__kernel", ["rbf", 'poly'])
        cs.add_hyperparameter(kernel)
        c = UniformFloatHyperparameter("SVC__C", 0.5, 200)
        cs.add_hyperparameter(c)
        use_svc_c = InCondition(child=kernel, parent=switch, values=["svc"])
        use_svc_kernel = InCondition(child=c, parent=switch, values=["svc"])

        criterion = CategoricalHyperparameter("RandomForestClassifier__criterion", ['gini', 'entropy'])
        cs.add_hyperparameter(criterion)
        minsplit = UniformIntegerHyperparameter("RandomForestClassifier__min_samples_split", 2, 4)
        cs.add_hyperparameter(minsplit)

        use_rf_crit = InCondition(child=criterion, parent=switch, values=["rf"])
        use_rf_minsplit = InCondition(child=minsplit, parent=switch, values=["rf"])

        cs.add_conditions([use_svc_c, use_svc_kernel, use_rf_crit, use_rf_minsplit])

        # Scenario object
        scenario = Scenario({"run_obj": "quality",
                             "cs": cs,
                             "deterministic": "true",
                             "wallclock_limit": self.time_limit,
                             "limit_resources": False,
                             'abort_on_first_run_crash': False
                             })

        # Optimize, using a SMAC directly
        smac = SMAC4BO(scenario=scenario, rng=42, tae_runner=self.objective_function_switch)
        _ = smac.optimize()

        runhistory_photon = [x.metrics_test[0].value for x in self.pipe.results.outer_folds[0].tested_config_list]
        runhistory_original = [1 - x for x in list(smac.solver.runhistory._cost_per_config.values())]

        min_len = min(len(runhistory_original), len(runhistory_photon))
        np.testing.assert_almost_equal(runhistory_photon[:min_len], runhistory_original[:min_len], 1)

    def objective_function_switch(self, cfg):
        cfg = {k: cfg[k] for k in cfg if cfg[k]}
        values = []

        train_indices = list(self.pipe.cross_validation.outer_folds.values())[0].train_indices
        self._validation_X, self._validation_y, _ = PhotonDataHelper.split_data(self.X, self.y,
                                                                                kwargs=None,
                                                                                indices=train_indices)

        switch = cfg["Estimator_switch"]
        del cfg["Estimator_switch"]
        for inner_fold in list(list(self.pipe.cross_validation.inner_folds.values())[0].values()):
            sc = PipelineElement("StandardScaler", {})
            pca = PipelineElement("PCA", {}, random_state=42)
            if switch == 'svc':
                est = PipelineElement("SVC", {}, random_state=42, gamma='auto')
                name = 'SVC'
            else:
                est = PipelineElement("RandomForestClassifier", {}, random_state=42)
                name = "RandomForestClassifier"
            my_pipe = PhotonPipeline([('StandardScaler', sc), ('PCA', pca), (name, est)])
            my_pipe.set_params(**cfg)
            my_pipe.fit(self._validation_X[inner_fold.train_indices, :],
                        self._validation_y[inner_fold.train_indices])
            values.append(accuracy_score(self._validation_y[inner_fold.test_indices],
                                         my_pipe.predict(self._validation_X[inner_fold.test_indices, :])
                                         )
                          )
        return 1 - np.mean(values)

    def test_facade(self):
        config_space = ConfigurationSpace()
        n_components = UniformIntegerHyperparameter("PCA__n_components", 5, 30)
        config_space.add_hyperparameter(n_components)
        scenario_dict = {"run_obj": "quality",
                         "deterministic": "true",
                         "cs": config_space,
                         "wallclock_limit": 60
                         }

        with self.assertRaises(ValueError):
            SMACOptimizer(facade="SMAC4BOO", scenario_dict=scenario_dict)

        with self.assertRaises(ValueError):
            facade = SMAC4BO(scenario=Scenario(scenario_dict))
            SMACOptimizer(facade=facade, scenario_dict=scenario_dict)

        facades = ["SMAC4BO", SMAC4BO, "SMAC4AC", SMAC4AC, "SMAC4HPO", SMAC4HPO, "BOHB4HPO", BOHB4HPO]
        for facade in facades:
            SMACOptimizer(facade=facade, scenario_dict=scenario_dict)

    def test_other(self):
        opt = SMACOptimizer(facade="SMAC4BO", intensifier_kwargs={'min_chall': 2})
        self.assertIsNotNone(opt.intensifier_kwargs)

        pipeline_elements = [PipelineElement('SVC', hyperparameters={'kernel': Categorical(["rbf", 'poly', "sigmoid"]),
                                                                     'C': [0.6]})]

        def of(x):
            return x ** 2
        with warnings.catch_warnings(record=True) as w:
            opt.prepare(pipeline_elements=pipeline_elements, maximize_metric=True, objective_function=of)
            assert any("PHOTONAI has detected some" in s for s in [e.message.args[0] for e in w])

        pipeline_elements = [PipelineElement("SVC", hyperparameters={'C': FloatRange(0.1, 0.5, range_type='geomspace'),
                                                                     'kernel': ['rbf']})]
        opt = SMACOptimizer(facade="SMAC4BO")
        with self.assertRaises(NotImplementedError):
            opt.prepare(pipeline_elements=pipeline_elements, maximize_metric=True, objective_function=of)

        pipeline_elements = [PipelineElement("SVC", hyperparameters={'C': NumberRange(1, 3, range_type='range')})]
        opt = SMACOptimizer(facade="SMAC4BO")
        with self.assertRaises(ValueError):
            opt.prepare(pipeline_elements=pipeline_elements, maximize_metric=True, objective_function=of)

    @classmethod
    def tearDownClass(cls) -> None:
        dirs = glob.glob("./smac3-output*/")
        for dir_name in dirs:
            rmtree(dir_name, ignore_errors=True)



from sklearn.base import BaseEstimator
from sklearn.preprocessing import OrdinalEncoder
from photonai.helper.helper import PhotonDataHelper
import numpy as np


class FeatureEncoder(BaseEstimator):

    def __init__(self):
        self.encoder_list = []

    def fit(self, X, y=None, **kwargs):
        # iterate X and if its a string column: encode and save encoder
        self.encoder_list = list()
        for i in range(X.shape[1]):
            feature = X[:, i]
            if isinstance(feature[0], str):
                new_encoder = OrdinalEncoder()
                feature = np.reshape(feature, (-1, 1))
                new_encoder.fit(feature)
                self.encoder_list.append(new_encoder)
            else:
                self.encoder_list.append(None)

    def transform(self, X, y=None, **kwargs):
        # iterate X and apply encoder if necessary
        new_X = None
        for i in range(X.shape[1]):
            feature = X[:, i]
            transformer = self.encoder_list[i]
            if transformer is not None:
                feature = np.reshape(feature, (-1, 1))
                trans_X = transformer.transform(feature)
            else:
                trans_X = feature
            new_X = PhotonDataHelper.stack_data_horizontally(new_X, trans_X)
        return new_X

    def fit_transform(self, X, y=None, **kwargs):
        self.fit(X, y)
        return self.transform(X, y)

    def inverse_transform(self, X, y=None, **kwargs):
        new_X = None
        for i in range(X.shape[1]):
            feature = X[:, i]
            transformer = self.encoder_list[i]
            if transformer is not None:
                feature = np.reshape(feature, (-1, 1))
                trans_X = transformer.inverse_transform(feature)
            else:
                trans_X = feature
            new_X = PhotonDataHelper.stack_data_horizontally(new_X, trans_X)
        return new_X



import os

import numpy as np
from sklearn.model_selection import KFold

from photonai.base import PhotonRegistry
from photonai.base import PipelineElement, Hyperpipe, OutputSettings
from photonai.helper.photon_base_test import PhotonBaseTest


class RegistryTest(PhotonBaseTest):


    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(RegistryTest, cls).setUpClass()

    def setUp(self):
        super(RegistryTest, self).setUp()
        self.test_directory = os.path.join(os.path.dirname(os.path.abspath(__file__)).split('test')[0], 'test')
        self.custom_folder = os.path.join(self.test_directory, 'base_tests/custom_elements/')
        self.registry = PhotonRegistry(self.custom_folder)

    def tearDown(self):
        super(RegistryTest, self).tearDown()
        if os.path.isfile(os.path.join(self.custom_folder, 'CustomElements.json')):
            os.remove(os.path.join(self.custom_folder, 'CustomElements.json'))

    def test_register_without_custom_folder(self):
        registry = PhotonRegistry()
        with self.assertRaises(ValueError):
            registry.register('SomeName', 'not_existing_file.SomeName', 'Estimator')

    def test_python_file_not_in_custom_folder(self):
        with self.assertRaises(FileNotFoundError):
            self.registry.register('SomeName', 'not_existing_file.SomeName', 'Estimator')

    def test_list_available_elements(self):
        self.registry.list_available_elements()
        self.registry.info('PCA')
        self.registry.info('NotExistingEstimator')

    def test_register_element(self):
        with self.assertRaises(ValueError):
            self.registry.register('MyCustomEstimator', 'custom_estimator.CustomEstimator', 'WrongType')

        self.registry.register('MyCustomEstimator', 'custom_estimator.CustomEstimator', 'Estimator')

        self.registry.activate()
        settings = OutputSettings(save_output=False)

        # DESIGN YOUR PIPELINE
        pipe = Hyperpipe('custom_estimator_pipe',
                         optimizer='random_grid_search',
                         optimizer_params={'n_configurations': 2},
                         metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                         best_config_metric='accuracy',
                         outer_cv=KFold(n_splits=2),
                         inner_cv=KFold(n_splits=2),
                         verbosity=1,
                         project_folder = './tmp/',
                         output_settings=settings)

        pipe += PipelineElement('MyCustomEstimator')

        pipe.fit(np.random.randn(30, 30), np.random.randint(0, 2, 30))

        self.registry.delete('MyCustomEstimator')

        os.remove(os.path.join(self.custom_folder, 'CustomElements.json'))

    def test_estimator_check_during_register(self):
        with self.assertRaises(NotImplementedError):
            self.registry.register('MyCustomEstimatorNoFit', 'custom_estimator.CustomEstimatorNoFit', 'Estimator')

        with self.assertRaises(NotImplementedError):
            self.registry.register('MyCustomEstimatorNoPredict', 'custom_estimator.CustomEstimatorNoPredict',
                                   'Estimator')

        with self.assertRaises(NotImplementedError):
            self.registry.register('MyCustomEstimatorNoEstimatorType',
                                   'custom_estimator.CustomEstimatorNoEstimatorType',
                                   'Estimator')

        with self.assertRaises(NotImplementedError):
            self.registry.register('MyCustomEstimatorNotReturningSelf',
                                   'custom_estimator.CustomEstimatorNotReturningSelf',
                                   'Estimator')

        e = self.registry.register('MyCustomEstimatorReturningFalsePredictions',
                                   'custom_estimator.CustomEstimatorReturningFalsePredictions',
                                   'Estimator')
        self.assertIsInstance(e, ValueError)

        e = self.registry.register('MyCustomEstimatorNotWorking', 'custom_estimator.CustomEstimatorNotWorking',
                                   'Estimator')

        self.assertIsInstance(e, ValueError)

        os.remove(os.path.join(self.custom_folder, 'CustomElements.json'))

    def test_transformer_needs_covariates(self):
        self.registry.register('MyCustomTransformerNeedsCovariates',
                               'custom_transformer.CustomTransformerNeedsCovariates',
                               'Transformer')
        with self.assertRaises(ValueError):
            self.registry.register('MyCustomTransformerNeedsCovariatesWrongInterface',
                                   'custom_transformer.CustomTransformerNeedsCovariatesWrongInterface',
                                   'Transformer')

    def test_transformer_needs_y(self):
        self.registry.register('MyCustomTransformerNeedsY',
                               'custom_transformer.CustomTransformerNeedsY',
                               'Transformer')

        with self.assertRaises(ValueError):
            self.registry.register('MyCustomTransformerNeedsYWrongInterface',
                                   'custom_transformer.CustomTransformerNeedsYWrongInterface',
                                   'Transformer')

    def test_add_module(self):
        json_module_file = os.path.join(self.custom_folder, "fake_module.json")

        # first add module
        self.registry.add_module(json_module_file)
        self.assertTrue("fake_module" in self.registry.PHOTON_REGISTRIES)

        # check if registered items are available
        self.assertTrue("FakeElement1" in self.registry.ELEMENT_DICTIONARY)
        fe = PipelineElement("FakeElement1")

        # then delete the module
        self.registry.delete_module("fake_module")
        self.assertFalse(os.path.isfile(os.path.join(self.registry.module_path, "fake_module.json")))

        self.assertFalse("FakeElement1" in self.registry.ELEMENT_DICTIONARY)
        with self.assertRaises(NameError):
            fe = PipelineElement("FakeElement1")

    def test_add_failing_module(self):
        json_module_file = os.path.join(self.custom_folder, "not_working_fake_module.json")

        # try to add module
        self.registry.add_module(json_module_file)

        # that should not have worked because we cant import first element
        self.assertFalse("FakeElement2001" in self.registry.ELEMENT_DICTIONARY)
        self.assertFalse(os.path.isfile(os.path.join(self.registry.module_path, "not_working_fake_module.json")))







import unittest

import numpy as np
from sklearn.model_selection import KFold, GroupShuffleSplit

from photonai.processing.photon_folds import FoldInfo


class PhotonFoldsTests(unittest.TestCase):

    def setUp(self):
        self.num_subjects = 100
        self.X = np.random.random((self.num_subjects, 45))
        self.y = np.random.randn(self.num_subjects)
        groups = []
        [groups.extend((np.ones((10,)) * i).tolist()) for i in range(10)]
        self.kwargs = {'groups': np.asarray(groups)}

    def base_assertions(self, cv, nr_of_folds, eval_final_performance=True):
        fold_list = FoldInfo.generate_folds(cv, self.X, self.y, self.kwargs,
                                            eval_final_performance=eval_final_performance)
        self.assertTrue(len(fold_list) == nr_of_folds)
        for generated_fold in fold_list:
            self.assertEqual(len(generated_fold.train_indices), (nr_of_folds - 1) * (self.num_subjects / nr_of_folds))
            self.assertEqual(len(generated_fold.test_indices), (self.num_subjects / nr_of_folds))
        # we always start with 1 for the investigator
        self.assertEqual(fold_list[0].fold_nr, 1)
        return fold_list

    def test_k_fold_generation(self):
        nr_of_folds = 5
        cv = KFold(n_splits=nr_of_folds)
        self.base_assertions(cv, nr_of_folds)

    def test_group_k_fold_generation(self):
        nr_of_folds = 5
        cv = GroupShuffleSplit(n_splits=nr_of_folds)
        fold_list = self.base_assertions(cv, nr_of_folds)

        # test that we have used the groups from the kwargs
        for generated_fold in fold_list:
            train_groups = np.unique(self.kwargs['groups'][generated_fold.train_indices])
            test_groups = np.unique(self.kwargs['groups'][generated_fold.test_indices])
            self.assertFalse(bool(set(train_groups) & set(test_groups)))

    def test_no_cv_strategy(self):
        test_size = 0.15
        fold_list = FoldInfo.generate_folds(None, self.X, self.y, self.kwargs, eval_final_performance=True,
                                            test_size=test_size)
        # check that we have only one outer fold, that is split in training and test according to test size
        self.assertEqual(len(fold_list), 1)
        self.assertEqual(len(fold_list[0].train_indices), (1 - test_size) * self.num_subjects)
        self.assertEqual(len(fold_list[0].test_indices), test_size * self.num_subjects)

    def test_no_cv_strategy_eval_final_performance_false(self):
        test_size = 0.15
        fold_list = FoldInfo.generate_folds(None, self.X, self.y, self.kwargs, eval_final_performance=False,
                                            test_size=test_size)
        # check that we have only one outer fold, that is split in training and test according to test size
        self.assertEqual(len(fold_list), 1)
        self.assertEqual(len(fold_list[0].train_indices), self.num_subjects)
        self.assertEqual(len(fold_list[0].test_indices), 0)

        fold_list = FoldInfo.generate_folds(None, range(len(self.y)), self.y, self.kwargs, eval_final_performance=False)
        self.assertEqual(len(fold_list), 1)
        self.assertTrue(np.array_equal(range(len(self.y)), fold_list[0].train_indices))

    def test_k_fold_generation_eval_final_performance_false(self):
        # this should change nothing
        nr_of_folds = 10
        cv = KFold(n_splits=nr_of_folds)
        self.base_assertions(cv, nr_of_folds, eval_final_performance=False)

    def test_data_overview(self):
        expected_outcome = {str(i): 10 for i in range(10)}
        data_count = FoldInfo.data_overview(self.kwargs['groups'].astype(int))
        self.assertDictEqual(expected_outcome, data_count)



from enum import Enum
import numpy as np
import numbers
import inspect
import warnings

from photonai.processing.metrics import Scorer
from photonai.photonlogger.logger import logger


class PhotonBaseConstraint:
    """Photon base constraint.

    The PHOTONAI base interface for any performance constraints
    that could speed up hyperparameter search.
    After a particular configuration is tested in one fold,
    the performance constraint objects are called to
    evaluate if the configuration is promising. If not,
    further testing in other folds is skipped to increase speed.

    """
    ENUM_STRATEGY = Enum("strategy", "first any mean")

    def __init__(self, strategy: str = 'first', metric: str = '', threshold: float = None, margin: float = 0, **kwargs):
        """
        Initialize the object.

        Parameters:
            strategy:
                One of [first, all, mean].

            metric:
                Name of metric to perform on.

            threshold:
                Limit for decision.

            margin:
                Value to be added for threshold value.
                Continue if:
                - Scorer: P(Config) > P(Dummy) + margin
                - Error:  P(Config) < P(Dummy) - margin

        """
        self._metric = None
        self._greater_is_better = None
        self._strategy = None

        # with setting property we automatically find greater_is_better
        self.metric = metric
        self.threshold = threshold
        self.margin = margin

        if not isinstance(self.margin, numbers.Number):
            msg = "Could not set margin in {}. Defaulting to 0.".format(str(type(self).__name__))
            logger.warning(msg)
            warnings.warn(msg)
            self.margin = 0
        self.strategy = strategy

    @property
    def strategy(self):
        return self._strategy

    @strategy.setter
    def strategy(self, value):
        try:
            self._strategy = PhotonBaseConstraint.ENUM_STRATEGY[value]
        except KeyError:
            raise KeyError("Your strategy: " + str(value) + " is not supported yet. Please use one of " +
                           str([x.name for x in PhotonBaseConstraint.ENUM_STRATEGY]))

    @property
    def metric(self):
        return self._metric

    @metric.setter
    def metric(self, value):
        try:
            self._metric = value
            self._greater_is_better = Scorer.greater_is_better_distinction(self._metric)
        except NameError:
            self._metric = "unknown"
            logger.warning("Your metric is not supported. Performance constraints are constantly False.")

    def shall_continue(self, config_item):
        """Function to evaluate if the constraint is reached.

        If it returns True, the testing of the configuration is continued.
        If it returns False, further testing of the configuration is skipped
        to increase speed of the hyperparameter search.

        Parameters:
            config_item:
                All performance metrics and other scoring information
                for all configuration's performance.
                Can be used to evaluate if the configuration has any potential
                to serve the model's learning task.

        """
        if self.metric == "unknown":
            msg = "The metric is not known. Please check the metric: {}. " \
                  "Performance constraints are constantly True.".format(self.metric)
            logger.warning(msg)
            warnings.warn(msg)
            return True
        if self.metric not in config_item.inner_folds[0].validation.metrics:
            msg = "The metric is not calculated. Please insert to Hyperpipe.metrics. " \
                  "Performance constraints are constantly False.".format(self.metric)
            logger.warning(msg)
            warnings.warn(msg)
            return False
        if self.threshold is None:
            msg = "Could not established threshold for performance constraint: {}. " \
                  "Continuing computation for all folds.".format(str(type(self).__name__))
            logger.warning(msg)
            warnings.warn(msg)
            return True

        if self._greater_is_better:
            if self.strategy.name == 'first':
                if config_item.inner_folds[0].validation.metrics[self.metric] < self.threshold:
                    return False
            elif self.strategy.name == 'any':
                if any(item < self.threshold for item in [x.validation.metrics[self.metric]
                                                          for x in config_item.inner_folds]):
                    return False
            elif self.strategy.name == 'mean':
                if np.mean([x.validation.metrics[self.metric] for x in config_item.inner_folds]) < self.threshold:
                    return False
            return True
        else:
            if self.strategy.name == 'first':
                if config_item.inner_folds[0].validation.metrics[self.metric] > self.threshold:
                    return False
            elif self.strategy.name == 'any':
                if any(item > self.threshold for item in [x.validation.metrics[self.metric]
                                                          for x in config_item.inner_folds]):
                    return False
            elif self.strategy.name == 'mean':
                if np.mean([x.validation.metrics[self.metric] for x in config_item.inner_folds]) > self.threshold:
                    return False
            return True

    def copy_me(self):
        new_me = type(self)(metric=self.metric)
        signature = inspect.getfullargspec(self.__init__)[0]
        for attr in signature:
            if not attr == 'self' and hasattr(self, attr) and attr != 'strategy':
                setattr(new_me, attr, getattr(self, attr))
            elif attr == 'strategy':
                setattr(new_me, attr, getattr(self, attr).name)
        return new_me


class MinimumPerformanceConstraint(PhotonBaseConstraint):
    """Minumum performance.

    Tests if a configuration performs better than a given limit for a particular metric.

    Examples:
        MinimumPerformance('accuracy', 0.96) tests if the configuration
        has at least a performance of 0.96 in
        (the) [first, all, mean] fold(s).
        If not further testing of the configuration is skipped,
        as it is regarded as not promising enough.

    """
    def __init__(self, metric: str = '', threshold: float = 1., strategy: str = 'first'):
        """
        Initialize the object.

        Parameters:
            metric:
                Name of metric to perform on.

            threshold:
                Limit for the decision.

            strategy:
                One of [first, all, mean].

        """
        super(MinimumPerformanceConstraint, self).__init__(strategy=strategy, metric=metric, threshold=threshold)


class DummyPerformanceConstraint(PhotonBaseConstraint):
    """
    Tests if a configuration performs better than a given limit for a particular metric.


    Examples:
        DummyPerformance('accuracy', 0.1) tests if the configuration has
        at least a 10% better performance than the dummy
        estimator. Distinguish between [first, all, mean] fold(s).
        If not further testing of the configuration is skipped,
        as it is regarded as not promising enough.

    """

    def __init__(self, metric: str = '', margin: float = 0, strategy: str = 'first'):
        """
        Initialize the object.

        Parameters:
            metric:
                Name of metric to perform on.

            margin:
                Value to be added for threshold value.
                Continue if:
                - Scorer: P(Config) > P(Dummy) + margin
                - Error:  P(Config) < P(Dummy) - margin

            strategy:
                One of [first, all, mean].

        """
        super(DummyPerformanceConstraint, self).__init__(strategy=strategy, metric=metric, margin=margin)

    def set_dummy_performance(self, dummy_result):
        """Set threshold with margin and dummy_performance value.
        At the time the object is created, this value is not known.
        """
        performance = dummy_result.validation.metrics[self.metric]
        if performance is None or performance is np.nan:
            raise ValueError("Could not set dummy threshold for metric " + self.metric)
        if self._greater_is_better:
            self.threshold = performance + self.margin
        else:
            self.threshold = performance - self.margin

    def copy_me(self):
        """Copy self object. Appending threshold to super.copy_me()."""
        new_me = super(DummyPerformanceConstraint, self).copy_me()
        if "threshold" in self.__dict__.keys():
            new_me.threshold = self.threshold
        return new_me


class BestPerformanceConstraint(PhotonBaseConstraint):
    """
    BestPerformance decides in every fold: challenger works better than incumbent
    true: eval next fold, false: eval next config
    better in all inner_folds: incumbent = challenger.

    """
    def __init__(self, metric: str = '', margin: float = 0, strategy: str = 'mean'):
        """
        Initialize the object.

        Parameters:
            metric:
                Name of metric to perform on.

            margin:
                Value to be added for threshold value.
                Continue if:
                - Scorer: P(Config) > P(Dummy) + margin
                - Error:  P(Config) < P(Dummy) - margin

            strategy:
                One of [first, all, mean].

        """
        super(BestPerformanceConstraint, self).__init__(strategy=strategy, margin=margin, metric=metric)

        self.threshold = None
        self.config_items = {}
        self.required_folds = 0
        self.run = 0

    def shall_continue(self, config_item):
        """Function to evaluate if the constraint is reached.

        If it returns True, the testing of the configuration is continued.
        If it returns False, further testing of the configuration is skipped
        to increase speed of the hyperparameter search.

        Parameters:
            config_item:
                All performance metrics and other scoring information
                for all configuration's performance.
                Can be used to evaluate if the configuration has any potential
                to serve the model's learning task.

        """
        self.required_folds = max(self.required_folds, len(self.eval_config_entries(config_item)))
        self.config_items[str(config_item.config_dict)] = self.eval_config_entries(config_item)

        # first run default 10 configs
        if self.run < 10 * self.required_folds - 1:
            self.run += 1
            return True

        # calc threshold for the first time over all available configs.
        if self.threshold is None:
            if self._greater_is_better:
                self.threshold = max([np.mean(x) for x in self.config_items.values()])
            else:
                self.threshold = min([np.mean(x) for x in self.config_items.values()])

        # calc std between inner_folds and average it over all configs [std only for multiple-run-configs]
        std = np.mean([np.std(x) for x in self.config_items.values() if len(self.config_items.values()) > 1])
        challenger = self.eval_config_entries(config_item)
        if self._greater_is_better:
            if np.mean(challenger) > self.threshold - std:
                if np.mean(challenger) > self.threshold and len(challenger) == self.required_folds:
                    self.threshold = np.mean(challenger)
                return True
        else:
            if np.mean(challenger) < self.threshold + std:
                if np.mean(challenger) < self.threshold and len(challenger) == self.required_folds:
                    self.threshold = np.mean(challenger)
                return True
        return False

    def copy_me(self):
        """Copy self object. Appending threshold to super.copy_me()."""
        new_me = super(BestPerformanceConstraint, self).copy_me()
        if "threshold" in self.__dict__.keys():
            new_me.threshold = self.threshold
        return new_me

    def eval_config_entries(self, config_item):
        return [x.validation.metrics[self.metric] for x in config_item.inner_folds]



import unittest
import numpy as np

from numpy.testing import assert_array_equal
from photonai.modelwrapper.label_encoder import LabelEncoder



class LabelEncoderTests(unittest.TestCase):

    def setUp(self):
        self.label_encoder = LabelEncoder()
        self.X = np.random.rand(5,5)
        self.values = np.array(["a", "b", "a", "a", "a", "c"])
        self.classes = np.unique(self.values)

    def test_fit(self):
        self.label_encoder.fit(self.X, self.values)
        assert_array_equal(self.label_encoder.classes_, self.classes)

    def test_transform(self):
        self.label_encoder.fit(self.X, self.values)
        trans_X, trans_y = self.label_encoder.transform(self.X, self.values)
        assert_array_equal(trans_y, [0, 1, 0, 0, 0, 2])

        with self.assertRaises(ValueError):
            self.label_encoder.transform(self.X, "unknown")

    def test_inverse_transform(self):
        self.label_encoder.fit(self.X, self.values)
        trans_X, trans_y = self.label_encoder.transform(self.X, self.values)
        assert_array_equal(self.label_encoder.inverse_transform(trans_y), self.values)

    def test_fit_transform(self):
        ret = self.label_encoder.fit_transform(self.X, self.values)
        assert_array_equal(ret, [0, 1, 0, 0, 0, 2])













import numpy as np
import statsmodels.api as sm
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import StandardScaler
from photonai.photonlogger.logger import logger

from typing import Union


class ConfounderRemoval(BaseEstimator, TransformerMixin):
    _estimator_type = "transformer"

    def __init__(self, standardize_covariates: bool = True, confounder_names: Union[str, list] = 'confounder'):

        self.needs_covariates = True
        self.needs_y = False
        self.standardize_covariates = standardize_covariates
        if isinstance(confounder_names, str):
            self.confounder_names = [confounder_names]
        else:
            self.confounder_names = confounder_names
        self.scalers = list()
        self.olsModel_params = None

    def _check_for_confounders(self, kwargs):
        if len(self.confounder_names) == 1:
            if self.confounder_names[0] in kwargs:
                sample_ols_confounder = kwargs[self.confounder_names[0]]
                return sample_ols_confounder
            else:
                raise KeyError("Variable 'confounder' not found in kwargs dictionary")
        elif len(self.confounder_names) > 1:
            sample_ols_confounder = list()
            len_list = list()
            for confound in self.confounder_names:
                if confound in kwargs:
                    sample_ols_confounder.append(kwargs[confound])
                    len_list.append(len(kwargs[confound]))
                else:
                    raise KeyError("Variable '{}' not found in kwargs dictionary".format(confound))
            if len(set(len_list)) != 1:
                raise ValueError("Provided confounders do not match in length")
            return np.stack(sample_ols_confounder, axis=1)

    def _validate_dimension(self, X, sample_ols_confounder):
        if X.shape[0] != sample_ols_confounder.shape[0]:
            err_msg = 'Number of samples (N=' + str(X.shape[0]) + ') is not the same as number of cases (N=' + str(
                sample_ols_confounder.shape[0]) + ') in confounder!'
            logger.error(err_msg)
            raise ValueError(err_msg)

    def _standardize(self, covariates, is_fit):
        logger.debug('Standardizing confounder prior to removal.')
        scaled_covs = list()
        if is_fit:
            # standardize covariates
            for cov in covariates.T:
                self.scalers.append(StandardScaler())
                scaled_covs.append(self.scalers[-1].fit_transform(cov.reshape(-1, 1)).squeeze())
            scaled_covs = np.asarray(scaled_covs).T
        else:
            for i, cov in enumerate(covariates.T):
                scaled_covs.append(self.scalers[i].transform(cov.reshape(-1, 1)).squeeze())
            scaled_covs = np.asarray(scaled_covs).T
        return scaled_covs

    def fit(self, X, y=None, **kwargs):

        sample_ols_confounder = self._check_for_confounders(kwargs)
        self._validate_dimension(X, sample_ols_confounder)

        # standardize covariates
        if self.standardize_covariates:
            sample_ols_confounder = self._standardize(sample_ols_confounder, is_fit=True)

        # sample_ols_confounder: confounder variables of the samples to be fitted
        ols_confounder = sm.add_constant(sample_ols_confounder, has_constant="add")
        self.olsModel_params = np.zeros(shape=(X.shape[1], ols_confounder.shape[1]))

        for i in range(X.shape[1]):
            ols_model = sm.OLS(endog=np.squeeze(X[:, i]), exog=ols_confounder).fit()
            self.olsModel_params[i] = ols_model.params
        return self

    def transform(self, X, y=None, **kwargs):
        logger.debug('Regress out confounder.')
        sample_ols_confounder = self._check_for_confounders(kwargs)
        self._validate_dimension(X, sample_ols_confounder)

        # standardize covariates
        if self.standardize_covariates:
            sample_ols_confounder = self._standardize(sample_ols_confounder, is_fit=False)

        sample_ols_confounder = sm.add_constant(sample_ols_confounder, has_constant="add")
        X_new = np.empty(X.shape)
        for i in range(X.shape[1]):
            preds = np.matmul(sample_ols_confounder, np.squeeze(self.olsModel_params[i]))
            residuum_feature_vector = np.squeeze(X[:, i]) - preds
            # residuum_feature_vector += self.olsModel_params[i, 0]  # add intercept
            X_new[:, i] = np.asarray(residuum_feature_vector)  # writing back the residuum of the feature vector
        return X_new, kwargs




from sklearn.base import BaseEstimator, RegressorMixin
from typing import Union


class RangeRestrictor(BaseEstimator, RegressorMixin):

    def __init__(self, low: Union[int, float] = 0, high: Union[int, float] = 100):
        self.low = low
        self.high = high
        self.needs_y = False
        self.needs_covariates = True

    def fit(self, X, y=None, **kwargs):
        return self

    def predict(self, X, **kwargs):
        X[X > self.high] = self.high
        X[X < self.low] = self.low
        return X



from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold
from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import FloatRange

my_pipe = Hyperpipe('basic_forest_pipe',
                    inner_cv=KFold(n_splits=2),
                    outer_cv=KFold(n_splits=2),
                    optimizer='sk_opt',
                    optimizer_params={'n_configurations': 15},
                    metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                    best_config_metric='accuracy',
                    project_folder='./tmp',
                    # this is how to make photonai calculate learning curves
                    # output and figures for this can be found in the project folder
                    learning_curves=True,
                    learning_curves_cut=FloatRange(0.1, 1, step=0.1))

my_pipe.add(PipelineElement('StandardScaler'))

my_pipe += PipelineElement('RandomForestClassifier')
X, y = load_breast_cancer(return_X_y=True)
my_pipe.fit(X, y)



import unittest
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold, StratifiedKFold

from photonai.base.json_transformer import JsonTransformer
from photonai.base import Hyperpipe, Preprocessing
from photonai.optimization import Categorical
from photonai.optimization.hyperparameters import IntegerRange, FloatRange
from photonai.base import Stack, Switch, Branch, PipelineElement
from test.base_tests.test_photon_elements import elements_to_dict


class JsonTransformerTest(unittest.TestCase):

    def setUp(self):
        """
        Set up for Scorer Tests.
        """
        self.maxDiff = None

    def test_branch_in_branch(self):
        """
        Test for deep Pipeline.
        """

        my_pipe = Hyperpipe('basic_stacking',
                            optimizer='grid_search',
                            metrics=['accuracy', 'precision', 'recall'],
                            best_config_metric='f1_score',
                            outer_cv=KFold(n_splits=2),
                            inner_cv=KFold(n_splits=3),
                            verbosity=1,
                            cache_folder="./cache/",
                            project_folder='./tmp/')

        # BRANCH WITH QUANTILTRANSFORMER AND DECISIONTREECLASSIFIER
        tree_qua_branch = Branch('tree_branch')
        tree_qua_branch += PipelineElement('QuantileTransformer')
        tree_qua_branch += PipelineElement('DecisionTreeClassifier',
                                           {'min_samples_split': IntegerRange(2, 4)},
                                           criterion='gini')

        # BRANCH WITH MinMaxScaler AND DecisionTreeClassifier
        svm_mima_branch = Branch('svm_branch')
        svm_mima_branch += PipelineElement('MinMaxScaler')
        svm_mima_branch += PipelineElement('SVC',
                                           {'kernel': ['rbf', 'linear'],  # Categorical(['rbf', 'linear']),
                                            'C': IntegerRange(0.01, 2.0)},
                                           gamma='auto')

        # BRANCH WITH StandardScaler AND KNeighborsClassifier
        knn_sta_branch = Branch('neighbour_branch')
        knn_sta_branch += PipelineElement('StandardScaler')
        knn_sta_branch += PipelineElement('KNeighborsClassifier')

        # voting = True to mean the result of every branch
        my_pipe += Stack('final_stack', [tree_qua_branch, svm_mima_branch, knn_sta_branch])
        my_pipe += PipelineElement('LogisticRegression', solver='lbfgs')

        json_transformer = JsonTransformer()
        pipe_json = json_transformer.create_json(my_pipe)
        my_pipe_reload = json_transformer.from_json(pipe_json)
        pipe_json_reload = pipe_json = json_transformer.create_json(my_pipe_reload)
        self.assertEqual(pipe_json, pipe_json_reload)

    def test_class_with_data_preproc(self):
        """
        Test for simple pipeline with data.
        """
        X, y = load_breast_cancer(return_X_y=True)

        # DESIGN YOUR PIPELINE
        my_pipe = Hyperpipe('basic_svm_pipe',
                            optimizer='grid_search',
                            metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                            best_config_metric='accuracy',
                            use_test_set=False,
                            outer_cv=KFold(n_splits=2),
                            inner_cv=KFold(n_splits=3),
                            verbosity=0,
                            random_seed=42)

        preprocessing = Preprocessing()
        preprocessing += PipelineElement("LabelEncoder")
        my_pipe += preprocessing

        # ADD ELEMENTS TO YOUR PIPELINE
        # first normalize all features
        my_pipe.add(PipelineElement('StandardScaler'))

        # then do feature selection using a PCA,
        my_pipe += PipelineElement('PCA',
                                   hyperparameters={'n_components': IntegerRange(10, 12)},
                                   test_disabled=True)

        # engage and optimize the good old SVM for Classification
        my_pipe += PipelineElement('SVC',
                                   hyperparameters={'kernel': Categorical(['rbf', 'linear'])
                                                    },
                                   C=2,
                                   gamma='scale')

        # NOW TRAIN YOUR PIPELINE
        my_pipe.fit(X, y)

        json_transformer = JsonTransformer()

        pipe_json = json_transformer.create_json(my_pipe)
        _ = elements_to_dict(my_pipe.copy_me())
        my_pipe_reload = json_transformer.from_json(pipe_json)
        pipe_json_reload = pipe_json = json_transformer.create_json(my_pipe_reload)

        self.assertEqual(pipe_json, pipe_json_reload)
        my_pipe_reload.fit(X, y)

        self.assertDictEqual(my_pipe.best_config, my_pipe_reload.best_config)

        self.assertDictEqual(elements_to_dict(my_pipe.copy_me()), elements_to_dict(my_pipe_reload.copy_me()))

    def test_class_with_data_02(self):
        """
        Test for Pipeline with data.
        """
        # DESIGN YOUR PIPELINE
        my_pipe = Hyperpipe(name='Estimator_pipe',
                            optimizer='grid_search',
                            metrics=['balanced_accuracy'],
                            best_config_metric='balanced_accuracy',
                            outer_cv=StratifiedKFold(n_splits=2, shuffle=True, random_state=42),
                            inner_cv=StratifiedKFold(n_splits=2, shuffle=True, random_state=42),
                            project_folder='./tmp/',
                            random_seed=42)

        # ADD ELEMENTS TO YOUR PIPELINE
        # first normalize all features
        my_pipe += PipelineElement('StandardScaler')

        # some feature selection
        my_pipe += PipelineElement('LassoFeatureSelection',
                                   hyperparameters={'percentile': FloatRange(start=0.1, step=0.1, stop=0.7,
                                                                             range_type='range'),
                                                    'alpha': FloatRange(0.5, 1)},
                                   test_disabled=True)

        # add imbalanced group handling
        my_pipe += PipelineElement('ImbalancedDataTransformer', method_name='SMOTE', test_disabled=False)

        # setup estimator stack
        est_stack = Stack(name='classifier_stack')
        clf_list = ['RandomForestClassifier', 'LinearSVC', 'NuSVC', "SVC", "MLPClassifier",
                    "KNeighborsClassifier", "Lasso", "PassiveAggressiveClassifier", "LogisticRegression",
                    "Perceptron", "RidgeClassifier", "SGDClassifier", "GaussianProcessClassifier",
                    "AdaBoostClassifier", "BaggingClassifier", "GradientBoostingClassifier"]

        for clf in clf_list:
            est_stack += PipelineElement(clf)
        my_pipe += est_stack

        my_pipe += PipelineElement('PhotonVotingClassifier')

        json_transformer = JsonTransformer()
        pipe_json = json_transformer.create_json(my_pipe)
        my_pipe_reload = json_transformer.from_json(pipe_json)

        self.assertDictEqual(elements_to_dict(my_pipe.copy_me()), elements_to_dict(my_pipe_reload.copy_me()))

    def test_class_switch(self):
        """
        Test for Pipeline with data.
        """
        my_pipe = Hyperpipe('basic_switch_pipe',
                            optimizer='random_grid_search',
                            optimizer_params={'n_configurations': 15},
                            metrics=['accuracy', 'precision', 'recall'],
                            best_config_metric='accuracy',
                            outer_cv=KFold(n_splits=3),
                            inner_cv=KFold(n_splits=5),
                            verbosity=0,
                            project_folder='./tmp/')

        # Transformer Switch
        my_pipe += Switch('TransformerSwitch',
                          [PipelineElement('StandardScaler'),
                           PipelineElement('PCA', test_disabled=True)])

        # Estimator Switch
        svm = PipelineElement('SVC',
                              hyperparameters={'kernel': ['rbf', 'linear']})

        tree = PipelineElement('DecisionTreeClassifier',
                               hyperparameters={'min_samples_split': IntegerRange(2, 5),
                                                'min_samples_leaf': IntegerRange(1, 5),
                                                'criterion': ['gini', 'entropy']})

        my_pipe += Switch('EstimatorSwitch', [svm, tree])

        json_transformer = JsonTransformer()

        pipe_json = json_transformer.create_json(my_pipe)
        my_pipe_reload = json_transformer.from_json(pipe_json)

        self.assertDictEqual(elements_to_dict(my_pipe.copy_me()), elements_to_dict(my_pipe_reload.copy_me()))



"""
Define custom metrics here
The method stub of all metrics is
function_name(y_true, y_pred)
"""
import warnings
import numpy as np
from typing import Union, Type, Callable, Optional, Tuple, Dict
from scipy.stats import spearmanr
from sklearn.metrics import accuracy_score

from photonai.photonlogger.logger import logger


class Scorer(object):
    """Scorer.

    Transforms a string literal into an callable instance of a particular metric.

    """

    ELEMENT_DICTIONARY: Dict[str, Tuple[str, str, str]] = {
        # Classification
        'matthews_corrcoef': ('sklearn.metrics', 'matthews_corrcoef', 'score'),
        'accuracy': ('sklearn.metrics', 'accuracy_score', 'score'),
        'f1_score': ('sklearn.metrics', 'f1_score', 'score'),
        'hamming_loss': ('sklearn.metrics', 'hamming_loss', 'error'),
        'log_loss': ('sklearn.metrics', 'log_loss', 'error'),
        'precision': ('sklearn.metrics', 'precision_score', 'score'),
        'recall': ('sklearn.metrics', 'recall_score', 'score'),
        'auc': ('sklearn.metrics', 'roc_auc_score', 'score'),
        'sensitivity': ('photonai.processing.metrics', 'sensitivity', 'score'),
        'specificity': ('photonai.processing.metrics', 'specificity', 'score'),
        'balanced_accuracy': ('photonai.processing.metrics', 'balanced_accuracy', 'score'),
        'categorical_accuracy': ('photonai.processing.metrics', 'categorical_accuracy_score', 'score'),

        # Regression
        'mean_squared_error': ('sklearn.metrics', 'mean_squared_error', 'error'),
        'mean_absolute_error': ('sklearn.metrics', 'mean_absolute_error', 'error'),
        'explained_variance': ('sklearn.metrics', 'explained_variance_score', 'score'),
        'r2': ('sklearn.metrics', 'r2_score', 'score'),
        'pearson_correlation': ('photonai.processing.metrics', 'pearson_correlation', 'score'),
        'spearman_correlation': ('photonai.processing.metrics', 'spearman_correlation', 'score'),
        'variance_explained':  ('photonai.processing.metrics', 'variance_explained_score', 'score')
    }

    CUSTOM_ELEMENT_DICTIONARY: Dict[str, Callable] = {}

    Metric_Type = Union[
        Callable,
        'keras.metrics.Metric',
        Type['keras.metrics.Metric']
    ]

    dynamic_keras_import = None

    @classmethod
    def try_import_keras(cls):
        try:
            cls.dynamic_keras_import = __import__('keras')
        except ImportError:
            pass

    @classmethod
    def register_custom_metric(cls, metric: Union[Metric_Type, Tuple[str, Metric_Type]]) -> Optional[str]:
        if cls.dynamic_keras_import is None:
            cls.try_import_keras()
        # if metric is already a string, don't do anything
        if isinstance(metric, str):
            return metric

        # derive name from metric class unless it is explicitly given with a tuple
        if metric is not None:
            if isinstance(metric, Tuple):
                metric_name = metric[0]
                metric = metric[1]
            elif cls.dynamic_keras_import is not None and isinstance(metric, cls.dynamic_keras_import.metrics.Metric):
                metric_name = "custom_" + str(metric.__module__) + '.' + str(type(metric).__name__)
            else:
                metric_name = "custom_" + str(metric.__module__) + '.' + str(metric.__name__)
            metric_name = metric_name.lower()
        else:
            raise ValueError("Metric is None")

        # Check if metric_name is already registered
        if metric_name in Scorer.CUSTOM_ELEMENT_DICTIONARY:
            warn_text = 'Custom metric name ' + metric_name + ' is ambiguous. Please specify metric as tuple with ' + \
                        'cooresponding name (e.g. instead of metrics=[keras.metrics.Accuracy] use ' \
                        'metrics=[(\'MetricName1\', keras.metrics.Accuracy)]. Only the first occurance of this ' \
                        'metric will be used!'
            logger.warning(warn_text)
            warnings.warn(warn_text)
            return None

        # derive a metric function from the given object
        if cls.dynamic_keras_import is not None and (
                (isinstance(metric, type) and issubclass(metric, cls.dynamic_keras_import.metrics.Metric))
                or isinstance(metric, cls.dynamic_keras_import.metrics.Metric)
        ):
            if isinstance(metric, type) and issubclass(metric, cls.dynamic_keras_import.metrics.Metric):
                metric_obj = metric()
            else:
                metric_obj = metric

            def metric_func(y_true, y_pred):
                metric_obj.reset_states()
                metric_obj.update_state(y_true=y_true, y_pred=y_pred)
                return float(cls.dynamic_keras_import.backend.eval(metric_obj.result()))

            Scorer.CUSTOM_ELEMENT_DICTIONARY[metric_name] = metric_func
        elif callable(metric):
            Scorer.CUSTOM_ELEMENT_DICTIONARY[metric_name] = metric
        return metric_name
    
    @classmethod
    def create(cls, metric: str) -> Optional[Callable]:
        """Searches for the metric by name and instantiates the according calculation function

        Parameters
        ----------
        metric: str
            The name of the metric as encoded in the ELEMENT_DICTIONARY.

        Returns
        -------
        metric_function: Callable
            A callable instance of the metric calculation.

        """
        if metric in Scorer.ELEMENT_DICTIONARY:
            try:
                desired_class_info = Scorer.ELEMENT_DICTIONARY[metric]
                desired_class_home = desired_class_info[0]
                desired_class_name = desired_class_info[1]
                imported_module = __import__(desired_class_home, globals(),
                                             locals(), desired_class_name, 0)
                desired_class = getattr(imported_module, desired_class_name)
                scoring_method = desired_class
                return scoring_method
            except AttributeError as ae:
                msg = 'Could not find according class: ' + Scorer.ELEMENT_DICTIONARY[metric]
                logger.error(msg)
                raise AttributeError(msg)
        elif metric in Scorer.CUSTOM_ELEMENT_DICTIONARY:
            return Scorer.CUSTOM_ELEMENT_DICTIONARY[metric]
        else:
            msg = 'Metric not supported right now:' + str(metric)
            logger.error(msg)
            raise NameError(msg)

    @staticmethod
    def greater_is_better_distinction(metric: str) -> bool:
        if metric in Scorer.ELEMENT_DICTIONARY:
            # for now do a simple hack and set greater_is_better
            # by looking at error/score in metric name
            specifier = Scorer.ELEMENT_DICTIONARY[metric][2]
            if specifier == 'score':
                return True
            elif specifier == 'error':
                return False
            else:
                # Todo: better error checking?
                error_msg = "Metric not suitable for optimizer."
                logger.error(error_msg)
                raise NameError(error_msg)
        elif metric in Scorer.CUSTOM_ELEMENT_DICTIONARY:
            # Check if it's a greater_is_better metric by calling it with example values
            metric_callable = Scorer.create(metric)
            y_true = [1, 0, 1, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0]
            y_pred = [0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 1, 1, 0]
            true_true = metric_callable(y_true, y_true)
            true_pred = metric_callable(y_true, y_pred)
            return true_true > true_pred
        else:
            logger.error('Specify valid metric to choose best config.')
        raise NameError('Specify valid metric to choose best config.')

    @staticmethod
    def calculate_metrics(y_true, y_pred, metrics):
        """Applies all metrics to the given predicted and true values.
        The metrics are encoded via a string literal which is mapped
        to the according calculation function.

        Parameters
        ----------
        y_true: list
            The truth values.
        y_pred: list
            The predicted values.
        metrics: list of str
            List of all metrics to be calculated from y_true and y_pred.

        Returns
        --------
        result: dict
            Dictionary with format name_of_metric -> value.

        """

        # Todo: HOW TO CHECK IF ITS REGRESSION?!
        # The following works only for classification
        # if np.ndim(y_pred) == 2:
        #     y_pred = one_hot_to_binary(y_pred)
        #     logger.warning("test_predictions was one hot encoded => transformed to binary")
        #
        # if np.ndim(y_true) == 2:
        #     y_true = one_hot_to_binary(y_true)
        #     logger.warning("test_y was one hot encoded => transformed to binary")

        output_metrics = {}
        if metrics:
            for metric in metrics:
                scorer = Scorer.create(metric)
                if scorer is not None:
                    scorer_value = scorer(y_true, y_pred)
                    output_metrics[metric] = scorer_value
                else:
                    output_metrics[metric] = np.nan

        return output_metrics


def one_hot_to_binary(one_hot_matrix):
    out = np.zeros((one_hot_matrix.shape[0]))
    for i in range(one_hot_matrix.shape[0]):
        out[i] = np.nonzero(one_hot_matrix[i, :])[0][0]
    return out


def categorical_accuracy_score(y_true, y_pred):
    if np.ndim(y_true) == 2:
        y_true = one_hot_to_binary(y_true)
    if np.ndim(y_pred) == 2:
        y_pred = one_hot_to_binary(y_pred)

    return accuracy_score(y_true, y_pred)


def pearson_correlation(y_true, y_pred):
    return np.corrcoef(y_true, y_pred)[0][1]


def spearman_correlation(y_true, y_pred):
    return spearmanr(y_true, y_pred)[0]


def variance_explained_score(y_true, y_pred):
    return np.square(pearson_correlation(y_true, y_pred))


def sensitivity(y_true, y_pred):  # = true positive rate, hit rate, recall
    if len(np.unique(y_true)) == 2:
        from sklearn.metrics import confusion_matrix
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
        return tp / (tp + fn)
    else:
        logger.info('Sensitivity (metric) is valid only for binary classification problems. You have ' +
                    str(len(np.unique(y_true))) + ' classes.')
        return np.nan


def specificity(y_true, y_pred):  # = true negative rate
    if len(np.unique(y_true)) == 2:
        from sklearn.metrics import confusion_matrix
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
        return tn / (tn + fp)
    else:
        logger.info('Specificity (metric) is valid only for binary classification problems. You have ' +
                    str(len(np.unique(y_true))) + ' classes.')
        return np.nan


def balanced_accuracy(y_true, y_pred):  # = true negative rate
    if len(np.unique(y_true)) == 2:
        return (specificity(y_true, y_pred) + sensitivity(y_true, y_pred)) / 2
    else:
        logger.info('Specificity (metric) is valid only for binary classification problems. You have ' +
                    str(len(np.unique(y_true))) + ' classes.')
        return np.nan



try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages


__version__ = '2.1.0'

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='photonai',
    packages=find_packages(),
    include_package_data=True,
    version=__version__,
    description="PHOTONAI is a high level python API for designing and optimizing machine learning pipelines.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license = "GPLv3",
    author='PHOTONAI Team',
    author_email='hahnt@wwu.de',
    url='https://www.photon-ai.com/',
    project_urls={
        "Source Code": "https://github.com/wwu-mmll/photonai/",
        "Documentation": "https://wwu-mmll.github.io/photonai/",
        "Bug Tracker": "https://github.com/wwu-mmll/photonai/issues",
    },
    download_url='https://pypi.org/project/photonai/#files',
    keywords=['machine learning', 'deep learning', 'neural networks', 'hyperparameter'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
    ],
    install_requires=[
        'numpy',
        'matplotlib',
        'scikit-learn',
        'keras',
        'pandas',
        'plotly',
        'imbalanced-learn',
        'pymodm',
        'scipy',
        'statsmodels',
        'prettytable',
        'seaborn',
        'joblib',
        'dask==2.30.0',
        'distributed==2.30.1',
        'scikit-optimize',
        'xlrd']
)



import os
from glob import glob
from pathlib import Path


def create_tests_example_script():
    root_path = Path(os.path.dirname(os.path.realpath(__file__))).parent
    examples_folder = root_path.joinpath('examples')
    examples_test_path = root_path.joinpath('test').joinpath('integration_tests')
    files = [f for f in glob(str(examples_folder) + "/**/*.py", recursive=True)]
    string = """

import warnings
from os.path import join, dirname, realpath
from pathlib import Path
from photonai.helper.photon_base_test import PhotonBaseTest
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)
warnings.filterwarnings("ignore", category=FutureWarning)


class TestRunExamples(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(TestRunExamples, cls).setUpClass()
    
    def setUp(self):     
        self.examples_folder = Path(dirname(realpath(__file__))).parent.parent.joinpath('examples')
    """

    for file in files:
        if not "__init__" in file:
            string += """
    def test_{}(self):
        exec(open(join(self.examples_folder, "{}")).read(), locals(), globals())
""".format(os.path.basename(file)[:-3], file.split('/examples/')[1])
    script = open(os.path.join(examples_test_path, "test_examples.py"), "w")
    script.write(string)
    script.close()


if __name__ == "__main__":
    create_tests_example_script()



from photonai.base import PipelineElement, Switch
from photonai.optimization import FloatRange, IntegerRange
from photonai.optimization.switch_optimizer.meta_optimizer import MetaHPOptimizer
from sklearn.datasets import load_breast_cancer
from ..grid_search_tests.test_grid_search import GridSearchOptimizerTest


class SwitchOptimizerTest(GridSearchOptimizerTest):

    def setUp(self):
        self.pipeline_elements = [PipelineElement("StandardScaler"),
                                  PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 20)}),
                                  Switch('estimators', [PipelineElement("SVC",
                                                                        {'C': FloatRange(0.1, 1, step=0.2)}),
                                                        PipelineElement('RandomForestClassifier',
                                                                        {'min_samples_split': IntegerRange(2, 5)})])
                                  ]
        self.optimizer = MetaHPOptimizer(name='random_grid_search')
        self.optimizer_name = "switch"
        self.num_configs = 3
        self.optimizer_params = {'name': 'sk_opt', 'n_configurations': self.num_configs}

    def test_wrong_setup(self):
        with self.assertRaises(ValueError):
            _ = MetaHPOptimizer(any_param_but_no_name=1)

        opt = MetaHPOptimizer(name='grid_search')
        with self.assertRaises(ValueError):
            opt.prepare([object()], True)

        opt = MetaHPOptimizer(name='random_name')
        with self.assertRaises(ValueError):
            opt.prepare([Switch('the_switch')], True)

    def test_one_opt_per_estimator(self):
        self.create_hyperpipe()
        for p in self.pipeline_elements:
            self.hyperpipe += p
        X, y = load_breast_cancer(True)
        self.hyperpipe.fit(X, y)

        # check there are three tested configs for each estimator
        self.assertEqual(len(self.hyperpipe.results.outer_folds[0].tested_config_list), 2*self.num_configs)

    def test_ask_advanced(self):
        pass
        #
        # with self.assertRaises(ValueError):
        #     super(SwitchOptimizerTest, self).test_ask_advanced()



""" PHOTONAI Classes for defining the hyperparameter search space and optimization strategies"""

from .hyperparameters import PhotonHyperparam, IntegerRange, FloatRange, Categorical, BooleanSwitch
from .performance_constraints import DummyPerformanceConstraint, MinimumPerformanceConstraint, BestPerformanceConstraint

from .grid_search.grid_search import GridSearchOptimizer, RandomGridSearchOptimizer
from .scikit_optimize.sk_opt import SkOptOptimizer
from .random_search.random_search import RandomSearchOptimizer
from .smac.smac import SMACOptimizer
from .nevergrad.nevergrad import NevergradOptimizer



from sklearn.datasets import fetch_openml
from sklearn.model_selection import KFold
from keras.callbacks import EarlyStopping

from photonai.base import Hyperpipe, PipelineElement

# WE USE THE BREAST CANCER SET FROM SKLEARN
X, y = fetch_openml('mnist_784', version=1, return_X_y=True)

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('basic_keras_multiclass_pipe',
                    optimizer='grid_search',
                    optimizer_params={},
                    metrics=['accuracy'],
                    best_config_metric='accuracy',
                    outer_cv=KFold(n_splits=2),
                    inner_cv=KFold(n_splits=2),
                    verbosity=1,
                    project_folder='./tmp/')

es = EarlyStopping(monitor='val_loss', mode='min', verbose=1, patience=10)

callbacks = [es]

# ADD ELEMENTS TO YOUR PIPELINE
my_pipe.add(PipelineElement('StandardScaler'))

# attention: shape of hidden_layer_sizes == shape of activations. If you want to choose a function in every layer,
# grid_search permits combinations with len(hidden_layer_size) != len(activations).
# Check out: hidden_layer_sizes=[25, 10], activations=['tanh', 'relu']
my_pipe += PipelineElement('KerasDnnClassifier',
                           hidden_layer_sizes=[10],
                           activations='relu',
                           nn_batch_size=128,
                           callbacks=callbacks,
                           epochs=50,
                           multi_class=True,
                           verbosity=1)

# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y)



import numpy as np
from sklearn.base import BaseEstimator


class BaseModelWrapper(BaseEstimator):
    """
    The PHOTONAI interface for implementing custom pipeline elements.

    PHOTONAI works on top of the scikit-learn object API,
    [see documentation](http://scikit-learn.org/stable/developers/contributing.html#apis-of-scikit-learn-objects).

    Your class should overwrite the following definitions:

    - `fit(data)`: learn or adjust to the data.

    If it is an estimator, which means it has the ability to learn,

    - it should implement `predict(data)`: using the learned model to generate prediction,
    - should inherit *sklearn.base.BaseEstimator* ([see here](
    http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html)),
    - inherits *get_params* and *set_params*.

    If it is an transformer, which means it preprocesses or prepares the data,

    - it should implement `transform(data)`: applying the logic to the data to transform it,
    - should inherit from *sklearn.base.TransformerMixin* ([see here](
    http://scikit-learn.org/stable/modules/generated/sklearn.base.TransformerMixin.html)),
    - inherits *fit_transform* as a concatenation of both fit and transform,
    - should inherit *sklearn.base.BaseEstimator* ([see here](
    http://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html))
    - inherits *get_params* and *set_params*.

    `Prepare for hyperparameter optimization`

    PHOTONAI expects a `definition for all parameters` you want to optimize in the hyperparameter search in the
    `constructor stub`, and to be addressable with the `same name as class variable`.
    In this way you can define any parameter and it is automatically prepared for the hyperparameter search process.

    See the [scikit-learn object API documentation](
    http://scikit-learn.org/stable/developers/contributing.html#apis-of-scikit-learn-objects)
    for more in depth information about the interface.

    """
    def __init__(self):
        pass

    def fit(self, data: np.ndarray, targets: np.ndarray = None):
        """
        Adjust the underlying model or method to the data.

        Parameters:
            data:
                The input samples of shape [n_samples, n_original_features].

            targets:
                The input targets of shape [n_samples, 1].

        Returns:
            IMPORTANT, must return self!

        """

    def predict(self, data: np.ndarray):
        """
        Use the learned model to make predictions.

        Parameters:
            data:
                The input samples of shape [n_samples, n_original_features].

        """

    def transform(self, data: np.ndarray, targets: np.ndarray = None):
        """
        Apply the method's logic to the data.

        Parameters:
            data:
                The input samples of shape [n_samples, n_original_features].

            targets:
                The input targets of shape [n_samples, 1]. Not necessary.

        """

    def get_params(self, deep: bool = True) -> dict:
        """
        Get the models parameters.
        Automatically implemented when inheriting from sklearn.base.BaseEstimator

        Parameters:
            deep:
                If True, will return the parameters for this estimator and contained subobjects that are estimators.

        Returns:
            Parameter names mapped to their values.
        """
        return super(BaseModelWrapper, self).get_params(deep=deep)

    def set_params(self, **kwargs):
        """
        Takes the given dictionary, with the keys being the variable name,
        and sets the object's parameters to the given values.
        Automatically implemented when inheriting from sklearn.base.BaseEstimator.

        Parameters:
            **kwargs: Estimator parameters.

        """
        super(BaseModelWrapper, self).set_params(**kwargs)



from sklearn.datasets import load_boston
from sklearn.model_selection import ShuffleSplit
import nevergrad as ng

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import BooleanSwitch, FloatRange

X, y = load_boston(return_X_y=True)

# list of all available nevergrad optimizer
print(list(ng.optimizers.registry.values()))

my_pipe = Hyperpipe('nevergrad_optimization_pipe',
                    optimizer='nevergrad',
                    optimizer_params={'facade': 'NGO', 'n_configurations': 30},
                    metrics=['mean_squared_error', 'pearson_correlation', 'mean_absolute_error', 'explained_variance'],
                    best_config_metric='mean_squared_error',
                    outer_cv=ShuffleSplit(n_splits=3, test_size=0.2),
                    inner_cv=ShuffleSplit(n_splits=3, test_size=0.2),
                    verbosity=0,
                    project_folder='./tmp/')

# ADD ELEMENTS TO YOUR PIPELINE
# first normalize all features
my_pipe += PipelineElement('StandardScaler')

my_pipe += PipelineElement('PCA', n_components='mle')

my_pipe += PipelineElement('Ridge', hyperparameters={'alpha': FloatRange(0.1, 100),
                                                     'fit_intercept': BooleanSwitch(),
                                                     'normalize': BooleanSwitch()})

my_pipe.fit(X, y)



import os
import shutil
import numpy as np
import joblib
import glob
from dask.distributed import Lock

from photonai.photonlogger.logger import logger


class CacheManager:

    __LOCK_STR = 'photon_cache_manager'

    def __init__(self, _hash=None, cache_folder=None, parallel_use: bool = False,
                 single_subject_caching: bool =False):
        self._hash = _hash
        self.cache_folder = cache_folder

        self.pipe_order = None
        self.cache_index = None
        self.state = None

        self.parallel_use = parallel_use
        self.single_subject_caching = single_subject_caching

    @property
    def hash(self):
        return self._hash

    @hash.setter
    def hash(self, value):
        if not isinstance(value, str):
            self._hash = str(value)
        else:
            self._hash = value

    class State:
        def __init__(self, config=None, nr_items=None,
                     first_data_hash=None, first_data_str: str = ''):
            self.config = config
            self.nr_items = nr_items
            self.first_data_hash = first_data_hash
            self.first_data_str = first_data_str

    def update_single_subject_state_info(self, X):
        self.state.first_data_hash = hash(str(X[0]))
        if isinstance(X[0], str):
            self.state.first_data_str = X[0]
        else:
            self.state.first_data_str = str(self.state.first_data_hash)

    def prepare(self, pipe_elements, config, X=None, single_subject_caching=False):

        self.read_cache_index()

        self.pipe_order = pipe_elements

        self.state = CacheManager.State(config=config)

        if X is not None:
            self.state.first_data_hash = hash(str(X[0]))
            if isinstance(X, np.ndarray):
                self.state.nr_items = X.shape[0]
            else:
                self.state.nr_items = len(X)
            self.state.first_data_str = str(self.state.first_data_hash)

        if single_subject_caching:
            self.state.nr_items = 1

    def _find_config_for_element(self, pipe_element_name):
        relevant_keys = list()
        for item in self.pipe_order:
            if item != pipe_element_name:
                relevant_keys.append(item)
            elif item == pipe_element_name:
                relevant_keys.append(item)
                break

        relevant_dict = dict()
        if self.state.config is not None and len(self.state.config) > 0:
            for key_name, key_value in self.state.config.items():
                key_name_list = key_name.split("__")
                if len(key_name_list) > 0:
                    item_name = key_name_list[0]
                else:
                    item_name = key_name

                if item_name in relevant_keys:
                    if isinstance(key_value, list):
                        key_value = frozenset(key_value)
                    relevant_dict[key_name] = key_value

        return hash(frozenset(relevant_dict.items()))

    def load_cached_data(self, pipe_element_name):

        cache_query = self.generate_cache_key(pipe_element_name)
        if cache_query in self.cache_index:
            if not self.single_subject_caching:
                logger.debug("Loading data from cache for " + pipe_element_name + ": "
                             + str(self.state.nr_items) + " items " + self.state.first_data_str
                             + " - " + str(self.state.config))
            filename = self.cache_index[cache_query]
            # lock = Lock(filename)
            # lock.acquire()
            with open(filename, 'rb') as f:
                (X, y, kwargs) = joblib.load(f)

            return X, y, kwargs
        return None

    def generate_cache_key(self, pipe_element_name):
        config_hash = self._find_config_for_element(pipe_element_name)
        cache_query = (pipe_element_name, self.hash, config_hash, self.state.nr_items, self.state.first_data_hash)
        return hash(cache_query)

    def check_cache(self, pipe_element_name):
        cache_query = self.generate_cache_key(pipe_element_name)

        if cache_query in self.cache_index:
            return True
        else:
            return False

    def save_data_to_cache(self, pipe_element_name, data):
        cache_query = self.generate_cache_key(pipe_element_name)
        filename = os.path.join(self.cache_folder, str(cache_query) + ".p")
        self.cache_index[cache_query] = filename
        if not self.single_subject_caching:
            logger.debug("Saving data to cache for " + pipe_element_name + ": " + str(self.state.nr_items) + " items "
                         + self.state.first_data_str + " - " + str(self.state.config))

        # write cached data to filesystem
        with open(filename, 'wb') as f:
            joblib.dump(data, f)

    def read_cache_index(self):
        cached_files = glob.glob(os.path.join(self.cache_folder, "*.p"))
        self.cache_index = {int(os.path.splitext(os.path.basename(i))[0]): i for i in cached_files}

    def clear_cache(self):
        CacheManager.clear_cache_files(self.cache_folder)

    @staticmethod
    def clear_cache_files(cache_folder, force_all=False):
        if cache_folder is not None:
            if os.path.isdir(cache_folder):
                for the_file in os.listdir(cache_folder):
                    file_path = os.path.join(cache_folder, the_file)
                    try:
                        if os.path.isfile(file_path):
                            os.unlink(file_path)
                        elif os.path.isdir(file_path):
                            if not file_path.endswith("DND") or force_all:
                                shutil.rmtree(file_path)
                    except Exception as e:
                        print(e)



import unittest
import numpy as np
import warnings

from sklearn.metrics import accuracy_score

from photonai.helper.helper import PhotonDataHelper
from photonai.base import PipelineElement, Hyperpipe
from photonai.base.photon_pipeline import PhotonPipeline
from photonai.optimization import FloatRange, Categorical, IntegerRange
from photonai.optimization.hyperparameters import NumberRange
from photonai.optimization import NevergradOptimizer
import photonai.optimization.nevergrad.nevergrad as photonai_ng

from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import ShuffleSplit

warnings.filterwarnings("ignore")

if photonai_ng.__found__:
    import nevergrad as ng


@unittest.skipIf(not photonai_ng.__found__, 'nevergrad not available')
class NevergradIntegrationTest(unittest.TestCase):

    def setUp(self):
        self.s_split = ShuffleSplit(n_splits=3, test_size=0.2, random_state=42)

        self.time_limit = 20

        # DESIGN YOUR PIPELINE
        self.pipe = Hyperpipe('basic_svm_pipe',
                              optimizer='nevergrad',
                              optimizer_params={'facade': ng.optimizers.NGO,
                                                'n_configurations': 20,
                                                'rng': 42},
                              metrics=['accuracy'],
                              best_config_metric='accuracy',
                              inner_cv=self.s_split,
                              verbosity=0,
                              project_folder='./tmp/')

    def simple_classification(self):
        dataset = load_breast_cancer()
        self.X = dataset["data"]
        self.y = dataset["target"]
        return self.X, self.y

    def test_photon_implementation_simple(self):
        """Integration test for simple pipeline without Switch."""
        # PHOTONAI implementation
        self.pipe.add(PipelineElement('StandardScaler'))
        self.pipe += PipelineElement('PCA', hyperparameters={'n_components': IntegerRange(5, 30)})
        self.pipe += PipelineElement('SVC', hyperparameters={'kernel': Categorical(["rbf", 'poly']),
                                                             'C': FloatRange(0.001, 2, range_type='logspace')},
                                     gamma='auto')
        self.X, self.y = self.simple_classification()
        self.pipe.fit(self.X, self.y)

        # direct AUTO ML implementation
        # Build Configuration Space which defines all parameters and their ranges
        cs = ng.p.Instrumentation(
            PCA__n_components=ng.p.Scalar(lower=5, upper=30).set_integer_casting(),
            SVC__kernel=ng.p.Choice(["rbf", 'poly']),
            SVC__C=ng.p.Log(lower=0.001, upper=2)
        )
        cs.random_state.seed(42)

        # Optimize, using a Nevergrad directly
        optimizer = ng.optimizers.NGO(parametrization=cs, budget=20)
        optimizer.minimize(self.objective_function_simple)

        runhistory_photon = [x.metrics_test[0].value for x in self.pipe.results.outer_folds[0].tested_config_list]
        runhistory_original = [1-x.mean for x in list(optimizer.optim.archive.bytesdict.values())]

        self.assertListEqual(runhistory_photon, runhistory_original)

    def objective_function_simple(self, **cfg):
        cfg = {k: cfg[k] for k in cfg if cfg[k]}
        values = []

        train_indices = list(self.pipe.cross_validation.outer_folds.values())[0].train_indices
        self._validation_X, self._validation_y, _ = PhotonDataHelper.split_data(self.X, self.y,
                                                                                kwargs=None, indices=train_indices)

        for inner_fold in list(list(self.pipe.cross_validation.inner_folds.values())[0].values()):
            sc = PipelineElement("StandardScaler", {})
            pca = PipelineElement("PCA", {}, random_state=42)
            svc = PipelineElement("SVC", {}, random_state=42, gamma='auto')
            my_pipe = PhotonPipeline([('StandardScaler', sc), ('PCA', pca), ('SVC', svc)])
            my_pipe.set_params(**cfg)
            my_pipe.fit(self._validation_X[inner_fold.train_indices, :],
                        self._validation_y[inner_fold.train_indices])
            values.append(accuracy_score(self._validation_y[inner_fold.test_indices],
                                         my_pipe.predict(self._validation_X[inner_fold.test_indices, :])
                                         )
                          )
        return 1-np.mean(values)

    def test_facade(self):
        facades = list(ng.optimizers.registry.keys())

        with self.assertRaises(ValueError):
            NevergradOptimizer(facade="Nevergrad4BOO", n_configurations=10)

        for facade in facades:
            NevergradOptimizer(facade=facade, n_configurations=10)

    def test_other(self):
        opt = NevergradOptimizer(facade="NGO", n_configurations=10)
        pipeline_elements = [PipelineElement('SVC', hyperparameters={'kernel': ["sigmoid", "rbf"],
                                                                     'C': [0.6], 'coef0': Categorical([0.5])})]
        of = lambda x: x ** 2
        with warnings.catch_warnings(record=True) as w:
            opt.prepare(pipeline_elements=pipeline_elements, maximize_metric=True, objective_function=of)
            assert any("PHOTONAI has detected some" in s for s in [e.message.args[0] for e in w])

        pipeline_elements = [PipelineElement("SVC", hyperparameters={'C': FloatRange(0.1, 0.5,
                                                                                     range_type='geomspace')})]
        opt = NevergradOptimizer(facade="NGO")
        with self.assertRaises(NotImplementedError):
            opt.prepare(pipeline_elements=pipeline_elements, maximize_metric=True, objective_function=of)

        pipeline_elements = [PipelineElement("SVC", hyperparameters={'C': NumberRange(1, 3, range_type='range')})]
        opt = NevergradOptimizer(facade="NGO")
        with self.assertRaises(ValueError):
            opt.prepare(pipeline_elements=pipeline_elements, maximize_metric=True, objective_function=of)



import unittest

from photonai.modelwrapper.base_model_wrapper import BaseModelWrapper


class BaseModelWrapperTest(unittest.TestCase):
    """
    Simple testclass for abstract BaseModelWrapper.
    Check required functions 'fit' and 'transform'.
    """

    def setUp(self):
        self.model_wrapper = BaseModelWrapper()

    def test_methods_available(self):
        if hasattr(self.model_wrapper, '_estimator_type'):
            if self.model_wrapper._estimator_type in ['classifier', 'regressor']:
                methods = ['fit', 'predict', 'get_params', 'set_params']
                for method in methods:
                    self.assertTrue(
                        (hasattr(self.model_wrapper, method) and callable(getattr(self.model_wrapper, method))))
            elif self.model_wrapper._estimator_type == 'transformer':
                methods = ['fit', 'transform', 'get_params', 'set_params']
                for method in methods:
                    self.assertTrue(
                        (hasattr(self.model_wrapper, method) and callable(getattr(self.model_wrapper, method))))



import warnings
import random
import numpy as np

from photonai.photonlogger.logger import logger


class PhotonHyperparam(object):
    """Photon hyperparameter.

    A base class that manages its own range of values.

    """
    def __init__(self, values: list):
        """
        Initialize the object.

        Parameters:
            values:
                Parameter Domain.

        """
        self.values = values

    def get_random_value(self, definite_list: bool = True):
        """
        Method for random search to get a random value based on the underlying domain.

        Parameters:
            definite_list:
                 Choice  between an element of a discrete list or a value within an interval.
                 In some cases, certain settings such as the
                 step size may otherwise be lost.

        """
        if definite_list:
            return random.choice(self.values)
        else:
            msg = "The PHOTONAI hyperparam has no own random function."
            logger.error(msg)
            raise NotImplementedError(msg)

    def __str__(self):
        return str(self.__class__) + str(self.values)


class Categorical(PhotonHyperparam):
    """
    Class for defining a definite list of values.

    """
    def __init__(self, values: list):
        """
        Initialize the object.

        Parameters:
            values:
                Definite list of hyperparameter values.

        """
        super(Categorical, self).__init__(values)

    def __getitem__(self, item):
        return self.values.__getitem__(item)


class BooleanSwitch(PhotonHyperparam):
    """Boolean switch.

    Class for defining a boolean hyperparameter.
    Equivalent to Categorical([True, False]).

    """
    def __init__(self):
        """Initialize the object."""
        super(BooleanSwitch, self).__init__([True, False])


class NumberRange(PhotonHyperparam):
    """Number range.

    Class for easily creating a range of numbers to be tested in the optimization process.

    Notes:

        Before the values of the domain are available,
        it is mandatory to call the transform method.

    """
    def __init__(self, start: float, stop: float, range_type: str, step: int = None,
                 num: int = None, num_type: type = np.int64, **kwargs):
        """
        Initialize the object.

        Parameters:
            start:
                The start value for generating the lower bound.
                The resulting interval includes the value.

            stop:
                The stop value for generating the upper bound.

                - if range_type == "range":
                    The end value is not included in the interval (see documentation of numpy.arange).
                - if range_type == "linspace"
                    The end value is included in the interval,
                    unless endpoint is set to False (see documentation of numpy.linspace).
                - if range_type == "logspace"
                    The end value is included in the interval,
                    unless endpoint is set to False (see documentation of numpy.logspace).
                - if range_type == "geomspace"
                    The end value is included in the interval,
                    unless endpoint is set to False (see documentation of numpy.logspace).

            range_type:
                Which method to use for generating the number interval.
                Possible options:

                - "range": numpy.arange is used to generate a list
                    of values separated by the same step width.
                - "linspace": numpy.linspace is used to generate a certain
                    number of values between start and stop.
                - "logspace": numpy.logspace is used to generate a logarithmically
                    distributed range of a certain length.
                - "geomspace": numpy.geomspace is used to generate numbers spaced
                    evenly on a log scale (geometric progression).

            num_type:
                The underlying datatype of the values.
                For the inheriting class IntegerRange it is set to np.int32.
                For the inheriting class FloatRange it is set to np.float32.

            step:
                if range_type == 'range', the spacing between values.

            num:
                if range_type == 'linspace', range_type == 'logspace', or range_type == 'geomspace',
                the number of samples to generate.

            **kwargs:
                Further parameters that should be passed to the numpy function chosen with range_type.

        """
        super(NumberRange, self).__init__([])
        self.start = start
        self.stop = stop
        self._range_type = None
        self.range_type = range_type
        self.range_params = kwargs
        self.num_type = num_type
        self.step = step
        self.num = num

    def transform(self):
        """
        Translates the definition into an area with values.
        These values are again stored in the attribute self.values.

        """
        if self.range_type == "geomspace" and self.start == 0:
            error_message = "Geometric sequence cannot include zero"
            logger.error(error_message)
            raise ValueError(error_message)
        if self.range_type == "range" and self.start > self.stop:
            warn_message = "NumberRange or one of its subclasses is empty cause np.arange " + \
                           "does not deal with start greater than stop."
            logger.warning(warn_message)
            warnings.warn(warn_message)

        values = []

        if self.range_type == "range":
            if not self.step:
                values = np.arange(self.start, self.stop, dtype=self.num_type, **self.range_params)
            else:
                values = np.arange(self.start, self.stop, self.step, dtype=self.num_type, **self.range_params)
        elif self.range_type == "linspace":
            if self.num:
                values = np.linspace(self.start, self.stop, num=self.num, dtype=self.num_type, **self.range_params)
            else:
                values = np.linspace(self.start, self.stop, dtype=self.num_type, **self.range_params)
        elif self.range_type == "logspace":
            if self.num_type == np.int32:
                raise ValueError("Cannot use logspace for integer,  use geomspace instead.")
            if self.num:
                values = np.logspace(self.start, self.stop, num=self.num, dtype=self.num_type, **self.range_params)
            else:
                values = np.logspace(self.start, self.stop, dtype=self.num_type, **self.range_params)
        elif self.range_type == "geomspace":
            if self.num:
                values = np.geomspace(self.start, self.stop, num=self.num, dtype=self.num_type, **self.range_params)
            else:
                values = np.geomspace(self.start, self.stop, dtype=self.num_type, **self.range_params)

        # convert to python datatype because mongodb needs it
        try:
            self.values = [values[i].item() for i in range(len(values))]
        except:
            msg = "PHOTON can not guarantee full mongodb support since you chose a non [np.integer, np.floating] " \
                  "subtype in NumberType.dtype."
            logger.warning(msg)
            warnings.warn(msg)
            self.values = values

    @property
    def range_type(self):
        return self._range_type

    @range_type.setter
    def range_type(self, value):
        range_types = ["range", "linspace", "logspace", "geomspace"]
        if value in range_types:
            self._range_type = value
        else:
            raise ValueError("Subclass of NumberRange supports only "+str(range_types)+" as range_type, not " +
                             repr(value))


class IntegerRange(NumberRange):
    """Integer range.

    Class for easily creating a range of integers
    to be tested in optimization process.

    """
    def __init__(self, start: float, stop: float, range_type: str = 'range',
                 step: int = None, num: int = None, **kwargs):
        """
        Initialize the object.

        Parameters:
            start:
                The start value for generating the lower bound.
                The resulting interval includes the value.

            stop:
                The stop value for generating the upper bound.

                - if range_type == "range":
                    The end value is not included in the interval (see documentation of numpy.arange).
                - if range_type == "linspace"
                    The end value is included in the interval,
                    unless endpoint is set to False (see documentation of numpy.linspace).
                - if range_type == "logspace"
                    The end value is included in the interval,
                    unless endpoint is set to False (see documentation of numpy.logspace).
                - if range_type == "geomspace"
                    The end value is included in the interval,
                    unless endpoint is set to False (see documentation of numpy.logspace).

            range_type:
                Which method to use for generating the number interval.
                Possible options,

                - "range": numpy.arange is used to generate a list
                    of values separated by the same step width.
                - "linspace": numpy.linspace is used to generate a certain
                    number of values between start and stop.
                - "logspace": numpy.logspace is used to generate a logarithmically
                    distributed range of a certain length.
                - "geomspace": numpy.geomspace is used to generate numbers spaced
                    evenly on a log scale (geometric progression).

            step:
                If range_type == 'range', the spacing between values.

            num:
                If range_type == 'linspace', range_type == 'logspace', or range_type == 'geomspace',
                the number of samples to generate.

            **kwargs:
                Further parameters that should be passed to the numpy function chosen with range_type.

        """
        super().__init__(start, stop, range_type, step, num, np.int32, **kwargs)

    def get_random_value(self, definite_list: bool = False):
        """
        Method for random search to get a random value based on the underlying domain.

        Parameters:
            definite_list:
                 Choice  between an element of a discrete list or a value within an interval.
                 As example, the step parameter would vanishes when this parameter
                 is set to False.

        """
        if definite_list:
            if not self.values:
                msg = "No values were set. Please use transform method."
                logger.error(msg)
                raise ValueError(msg)
            return random.choice(self.values)
        else:
            return random.randint(self.start, self.stop-1)


class FloatRange(NumberRange):
    """Float range.

    Class for easily creating an interval of numbers
    to be tested in the optimization process.

    """
    def __init__(self, start: float, stop: float, range_type: str = 'linspace',
                 step: float = None, num: int = None, **kwargs):
        """
        Initialize the object.

        Parameters:
            start:
                The start value for generating the lower bound.
                The resulting interval includes the value.

            stop:
                The stop value for generating the upper bound.

                - if range_type == "range":
                    The end value is not included in the interval (see documentation of numpy.arange).
                - if range_type == "linspace"
                    The end value is included in the interval,
                    unless endpoint is set to False (see documentation of numpy.linspace).
                - if range_type == "logspace"
                    The end value is included in the interval,
                    unless endpoint is set to False (see documentation of numpy.logspace).
                - if range_type == "geomspace"
                    The end value is included in the interval,
                    unless endpoint is set to False (see documentation of numpy.logspace).

            range_type:
                Which method to use for generating the number interval.
                Possible options,

                - "range": numpy.arange is used to generate a list
                    of values separated by the same step width.
                - "linspace": numpy.linspace is used to generate a certain
                    number of values between start and stop.
                - "logspace": numpy.logspace is used to generate a logarithmically
                    distributed range of a certain length.
                - "geomspace": numpy.geomspace is used to generate numbers spaced
                    evenly on a log scale (geometric progression).

            step:
                If range_type == 'range', the spacing between values.

            num:
                If range_type == 'linspace', range_type == 'logspace', or range_type == 'geomspace',
                the number of samples to generate.

            kwargs:
                Further parameters that should be passed to the numpy function chosen with range_type.

        """
        super(FloatRange, self).__init__(start, stop, range_type, step, num, np.float64, **kwargs)

    def get_random_value(self, definite_list: bool = False):
        """
        Method for random search to get a random value based on the underlying domain.

        Parameters:
            definite_list:
                 Choice  between an element of a discrete list or a value within an interval.
                 As example, the num parameter would vanishes when this parameter
                 is set to False.

        """
        if definite_list:
            if not self.values:
                msg = "No values were set. Please use transform method."
                logger.error(msg)
                raise ValueError(msg)
            return random.choice(self.values)
        else:
            return random.uniform(self.start, self.stop)



import unittest
import numpy as np

from numpy.testing import assert_array_equal
from photonai.modelwrapper.OrdinalEncoder import FeatureEncoder


class FeatureEncoderTests(unittest.TestCase):

    def setUp(self):
        self.ordinal_encoder = FeatureEncoder()
        self.X = np.array([["a", 3, "b"], ["x", 5, "b"], ["a", 3, "c"]], dtype=object)
        self.values = np.array([1, 2, 1])
        self.classes = np.unique(self.values)

        self.intX = np.array([[1, 2, 3], [2, 3, 4], [3, 4, 5]])

        self.X_with_bool = np.array([[1, 2, True], [24, 5, False]], dtype=object)

    def test_fit(self):
        self.ordinal_encoder.fit(self.X, self.values)
        self.assertEqual(len(self.ordinal_encoder.encoder_list), 3)
        self.assertIsNone(self.ordinal_encoder.encoder_list[1])

    def test_transform(self):
        self.ordinal_encoder.fit(self.X, self.values)
        result = self.ordinal_encoder.transform(np.array([["a",1, "b"], ["a", 1, "b"], ["a", 1, "b"]],  dtype=object))
        assert_array_equal(result,  np.array([[0, 1, 0], [0, 1, 0], [0, 1, 0]], dtype=object))

    def test_inverse_transform(self):
        self.ordinal_encoder.fit(self.X, self.values)
        result = self.ordinal_encoder.inverse_transform(np.array([[0, 1, 0], [0, 1, 0], [0, 1, 0]], dtype=object))
        assert_array_equal(result, np.array([["a", 1, "b"], ["a", 1, "b"], ["a", 1, "b"]], dtype=object))

    def test_fit_transform(self):
        result = self.ordinal_encoder.fit_transform(self.X, self.values)
        assert_array_equal(result, np.array([[0, 3, 0], [1, 5, 0], [0, 3, 1]], dtype=object))

        result = self.ordinal_encoder.fit_transform(self.intX)
        assert_array_equal(result, self.intX)

        result = self.ordinal_encoder.fit_transform(np.array([["a", 1, "b"], ["a", 1, "b"], ["a", 1, "b"]],
                                                             dtype=object))
        assert_array_equal(result, np.array([[0, 1, 0], [0, 1, 0], [0, 1, 0]], dtype=object))

    def only_strings(self):
        result = self.ordinal_encoder.fit_transform(self.X_with_bool)
        assert_array_equal(result, self.X_with_bool)




